"""QUERy for EXtraction"""

from xml.etree import ElementTree

from .querxml import FLOAT_SIZE, _set_query, _get_from_key_id


class MastersEx:
    def __init__(self, xml: ElementTree):
        self.xml = xml
        self.default_namespace = _set_query(self.xml, '')
        self._objects = self.findall("object")
        self._children = {
            self.findattr(object, "object_id"): self.findattr(object, "name")
            for object in self._objects
        }
        self._memberships = self.findall("membership")
        self._keys = self.findall("key")
        self._key_index = {
            (
                self.findattr(key_index, "key_id"),
                self.findattr(key_index, "period_type_id")
            ): (
                self.findattr(key_index, "position"),
                self.findattr(key_index, "length")
            ) for key_index in self.findall("key_index")
        }

    def find(self, query):
        return self.xml.find(f"{self.default_namespace}t_{query}")

    def findall(self, query):
        return self.xml.findall(f"{self.default_namespace}t_{query}")

    def findattr(self, leaf, attr):
        try:
            return leaf.find(f"{self.default_namespace}{attr}").text
        except AttributeError:
            raise AttributeError(f"{leaf} object has no attribute {attr}")

    def _find_id(self, attr, **kwargs):
        if hasattr(self, f"_{attr}s"):
            leaves = getattr(self, f"_{attr}s")
        else:
            leaves = self.findall(attr)
        for leaf in leaves:
            if all(
                self.findattr(leaf, filter_key) == filter_id
                for filter_key, filter_id in kwargs.items()
            ):
                return self.findattr(leaf, f"{attr}_id")

    def find_class_id(self, class_name: str):
        return self._find_id("class", name=class_name)

    def find_collection_id(self, collection_name: str):
        return self._find_id("collection", name=collection_name)

    def find_membership_id(self, parent_class_id, parent_id, child_class_id, child_id):
        return self._find_id(
            "membership",
            parent_object_id=parent_id,
            parent_class_id=parent_class_id,
            child_object_id=child_id,
            child_class_id=child_class_id,
        )

    def find_object_id(self, object_name, object_class_id):
        return self._find_id("object", name=object_name, class_id=object_class_id)

    def findall_membership_id(
        self,
        parent_class,
        parent,
        child_class,
        children,
    ):
        parent_class_id = self.find_class_id(parent_class)
        parent_id = self.find_object_id(parent, parent_class_id)
        child_class_id = self.find_class_id(child_class)
        for child in children:
            child_id = self.find_object_id(child, child_class_id)
            yield self.find_membership_id(parent_class_id, parent_id, child_class_id, child_id)

    def findall_key(self, **kwargs):
        # keys = []
        keys = {}
        for key in self._keys:
            if all(
                self.findattr(key, filter_key) == filter_id
                for filter_key, filter_id in kwargs.items()
            ) or not kwargs:
                prop_id = self.findattr(key, f"property_id")
                memb_id = self.findattr(key, f"membership_id")
                sample_id = self.findattr(key, f"sample_id")
                key_id = self.findattr(key, f"key_id")
                keys[(prop_id, memb_id, sample_id)] = key_id
                # keys.append(key)
        return keys

    def _find_key_id(self, kwargs, prefilter=None):
        key_ids = self._keys if prefilter is None else prefilter
        for key_id in key_ids:
            if all(
                self.findattr(key_id, filter_key) == filter_id
                for filter_key, filter_id in kwargs.items()
            ) or not kwargs:
                id_found = self.findattr(key_id, f"key_id")
                return id_found

    def find_key_index(self, prefilter, period_id, property_id, membership_id, sample_id):
        # key_id = self._find_key_id(kwargs, prefilter)
        key_id = prefilter[property_id, membership_id, sample_id]
        position, length = self._key_index[key_id, period_id]
        return int(position) // FLOAT_SIZE, int(length)


    def filter_children_from_parent(self, parent_class, parent, children_class, children):
        parent_class_id = self.find_class_id(parent_class)
        parent_id = self.find_object_id(parent, parent_class_id)
        children_class_id = self.find_class_id(children_class)
        for membership in self._memberships:
            membership_child_name = self._children[self.findattr(membership, "child_object_id")]
            if (
                self.findattr(membership, "parent_class_id") == parent_class_id
                and self.findattr(membership, "parent_object_id") == parent_id
                and self.findattr(membership, "child_class_id") == children_class_id
                and membership_child_name in children
            ):
                yield membership_child_name



import itertools
import logging
import mmap
import os
import tempfile
import time
from typing import Literal, TypeAlias
from xml.etree import ElementTree
from zipfile import ZipFile

import numpy as np
import pandas as pd
import polars as pl

from .kernel import (
    MastersEx, FLOAT_SIZE,
    query_horizon,
    query_classes, query_categories, query_children,
    query_phases, query_samples, query_samples_name,
    query_properties, query_property_id, query_property_unit,
    query_model_name, get_phase_id, get_period_id, get_period_type_id,
    query_timeslice_id
)
from .utils import QueryError


PivotType: TypeAlias = Literal["auto", "none", "property", "melt"]


def _get_period_type(period_type: str) -> str:
    period_type = period_type.lower()
    if period_type.startswith('i'):
        return "Interval"
    elif period_type.startswith('f') or period_type.startswith('y'):
        return "FiscalYear"
    raise TypeError(f"{period_type} is not a supported PeriodType!")


def _get_sample_name(sample_id: str) -> str:
    if statistic_name := {"-3": "Max", "-2": "Min", "-1": "Std", "0": "Mean"}.get(sample_id):
        return statistic_name
    elif sample_id.startswith('-'):
        return f"Statistic {sample_id[1:]}"
    return f"Sample {sample_id}"


def _get_xml_and_bin(zip_path: str) -> tuple:
    with ZipFile(zip_path) as zipsol:
        xml_name = next(filter(lambda filename: filename.endswith("Solution.xml"), zipsol.namelist()))
        with zipsol.open(xml_name) as xml_file:
            pointer = ElementTree.parse(xml_file)
        try:
            with zipsol.open("t_data_4.BIN") as bin_file:
                yearray = np.frombuffer(bin_file.read())
        except KeyError:
            yearray = None
        tmp_dir = tempfile.mkdtemp(prefix="DarkEx")
        try:
            bin_path = zipsol.extract("t_data_0.BIN", tmp_dir)
            logging.info(f"Temporary directory created: {tmp_dir}")
        except KeyError:
            os.rmdir(tmp_dir)
            bin_path = ""
    return pointer, bin_path, yearray


def _map_array(bin_path: str) -> tuple:
    with open(bin_path, "r+b") as memmap_bin:
        memmap = mmap.mmap(memmap_bin.fileno(), 0)
        array_length = len(memmap) // FLOAT_SIZE
        array_shape = (array_length,)
        array = np.ndarray(array_shape, buffer=memmap, dtype=float)
    return memmap, array


class DarkSol:
    def __init__(self, path: str):
        logging.basicConfig(level=logging.INFO, format="\x1b[35;20m%(levelname)s: %(message)s\x1b[0m")
        logging.debug("Loading solution...")
        start_time = time.time()
        self._pointer, self._bin, self._yearray = _get_xml_and_bin(path)
        self._kernel = MastersEx(self._pointer)
        if self._bin:
            self._memmap, self._array = _map_array(self._bin)
        else:
            self._memmap, self._array = None, None
        end_time = time.time()
        loading_time = round(end_time - start_time, 3)
        logging.info(f"Solution loaded in {loading_time} s, ready for extraction")
        self._horizon_type = None
        self._horizon = []
        self._is_closed = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self._is_closed or self._array is None:
            return
        self._memmap.close()
        os.remove(self._bin)
        os.rmdir(os.path.dirname(self._bin))
        logging.info(f"Temporary directory deleted: {self._bin}")
        self._is_closed = True

    def query_category(self, category_class: str) -> list:
        """Return the list of all available categories.

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_category()
        ["foo", "bar"]
        """
        categories = list(query_categories(self._pointer, category_class))
        if categories:
            categories.remove('-')
        return categories

    def query_class(self) -> list:
        """Return the list of all available classes.

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_class()
        ["Generator", "Reserve", "Region", "Node"]
        """
        return list(query_classes(self._pointer))

    def query_children(
            self,
            children_class: str,
            category: str = None,
            parent_class: str = "System",
            parent: str = "System"
    ) -> list:
        """Return the list of all children.

        Parameters
        __________
        children_class
            Which objects to query.

        category
            Filter children from a category listed in self.query_category(children_class).
            Default is None which means no filtering.

        parent_class
            Filter children by keeping only objects linked to another object from this class. Must be used with parent.
            Default is "System" which means no filtering.

        parent
            Filter children by keeping only objects linked to another object named parent.
            parent must be an object of class parent_class.
            Default is "System" which means no filtering.

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> # Assumption: each generator of this solution is named after its region followed by its category.
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_children("Generator")
        ["Limgrave foo", "Limgrave bar", "Caelid foo"]
        >>>
        >>>     ds.query_children("Generator", category="foo")
        ["Limgrave foo", "Caelid foo"]
        >>>
        >>>     ds.query_children("Generator", parent_class="Region", parent="Limgrave")
        ["Limgrave foo", "Limgrave bar"]
        """
        children = list(query_children(self._pointer, children_class, category))
        if children and parent_class != "System" and parent != "System":
            children = list(self._kernel.filter_children_from_parent(parent_class, parent, children_class, children))
        return children

    def query_model(self) -> str:
        """Return the name of the model.

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_model()
        "Open World"
        """
        return query_model_name(self._pointer)

    def query_phase(self) -> list:
        """Return the list of all available phases.

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_phase()
        ["LTPlan", "MTSchedule", "STSchedule"]
        """
        return list(query_phases(self._pointer))

    def query_property(self, children_class: str, parent_class: str = "System") -> list:
        """Return the list of all properties of a class.

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_property("Generator")
        ["Installed Capacity", "Generation"]
        >>>
        >>>     ds.query_property("Generator", "Reserve")
        ["Available Response", "Provision"]
        """
        return list(query_properties(self._pointer, children_class, parent_class))

    def query_sample(self, *, exclude_statistics: bool = False) -> list:
        """Return the list of all available samples, including the negative ones that corresponds to statistical values (e.g. "0" to "Mean").

        Parameters
        ----------
        .. versionadded:: 0.2.0: exclude_statistics (see below)
        exclude_statistics:
            Exclude negative samples

        Examples
        ________
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_sample()
        ["0", "1", "2", "3"]
        >>>
        >>>     ds.query_sample(exclude_statistics=True)
        ["1", "2", "3"]
        """
        samples = query_samples(self._pointer)
        if exclude_statistics:
            samples = filter(lambda sample: int(sample) > 0, samples)
        return list(samples)

    def query_sample_name(self, *, ignore_errors: bool = True) -> dict:
        """.. versionadded:: 0.2.0
        Return a dict that matches the sample with its name.

        Parameters
        ----------
        ignore_errors
            If the samples have no name, raise a QueryError if `ignore_errors` is False,
            or if `ignore_errors` is True, these names will be given (default behaviour):

            - 0 is "Mean", -1 is "Std", -2 is "Min", -3 is "Max"
            - Other negative samples will be named as "Statistic"
            - Positive samples will be named as "Sample"

        Example
        -------
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query_sample_name()
        {'0': 'Mean', '1': 'Sample 1', '2': 'Sample 2', '3': 'Sample 3'}
        """
        samples_id = self.query_sample()
        samples_name = query_samples_name(self._pointer)
        try:
            return dict(zip(samples_id, samples_name))
        except AttributeError:
            if ignore_errors:
                logging.warning("No sample name found; return based on fairly dark assumption.")
                samples_name = map(_get_sample_name, samples_id)
                return dict(zip(samples_id, samples_name))
            raise QueryError("Sample has no <sample_name> attribute.")

    def _format_extraction(
            self,
            array,
            prefilter,
            period_id,
            sample_id,
            child_name=None,
            membership_id=None,
            memberships={},
            property_name=None,
            property_id=None,
            properties={},
    ):
        result = {}
        if child_name is not None:
            result["Name"] = child_name
        if property_name is not None:
            result["Property"] = property_name
        result |= {
            "Sample": sample_id,
            "Datetime": self._horizon
        }
        size = len(self._horizon)
        # pivot == "none"
        for child, memb_id in memberships.items():
            try:
                address, size = self._kernel.find_key_index(
                    prefilter=prefilter,
                    period_id=period_id,
                    membership_id=memb_id,
                    sample_id=sample_id,
                    property_id=property_id,
                )
            except KeyError:
                raise QueryError(f"No data found for {property_name} of {child} with Sample {sample_id}")
            result |= {child: array[address:address + size].copy()}
        # pivot == "property"
        for prop, prop_id in properties.items():
            try:
                address, size = self._kernel.find_key_index(
                    prefilter=prefilter,
                    period_id=period_id,
                    membership_id=membership_id,
                    sample_id=sample_id,
                    property_id=prop_id,
                )
            except KeyError:
                raise QueryError(f"No data found for {prop} of {child_name} with Sample {sample_id}")
            result |= {prop: array[address:address + size].copy()}
        # pivot == "melt"
        if not memberships and not properties:
            try:
                address, size = self._kernel.find_key_index(
                    prefilter=prefilter,
                    period_id=period_id,
                    membership_id=membership_id,
                    sample_id=sample_id,
                    property_id=property_id,
                )
            except KeyError:
                raise QueryError(f"No data found for {property_name} of {child_name} with Sample {sample_id}")
            result |= {"Value": array[address:address + size].copy()}
        if skipped_hours := len(self._horizon) - size:
            # logging.info(f"Dark behaviour: last {skipped_hours} hours don't have values")
            result["Datetime"] = self._horizon[:size]
        return pl.DataFrame(result)

    def _process_extraction(
            self,
            array,
            prefilter,
            period_id,
            memberships,
            properties,
            samples,
            pivot,
    ) -> pl.DataFrame:
        if pivot == "auto":
            pivot = "property" if len(properties) >= len(memberships) else "none"
        if pivot == "property":
            return pl.concat((
                self._format_extraction(
                    array,
                    prefilter,
                    period_id,
                    sample_id,
                    child_name=child,
                    membership_id=membership_id,
                    properties=properties,
                )
                for sample_id, (child, membership_id)
                in itertools.product(samples, memberships.items())
            ))
        elif pivot == "none":
            return pl.concat((
                self._format_extraction(
                    array,
                    prefilter,
                    period_id,
                    sample_id,
                    memberships=memberships,
                    property_name=property_name,
                    property_id=property_id,
                )
                for sample_id, (property_name, property_id)
                in itertools.product(samples, properties.items())
            ))
        elif pivot == "melt":
            return pl.concat((
                self._format_extraction(
                    array,
                    prefilter,
                    period_id,
                    sample_id,
                    child_name=child,
                    membership_id=membership_id,
                    property_name=property_name,
                    property_id=property_id,
                )
                for sample_id, (property_name, property_id), (child, membership_id)
                in itertools.product(samples, properties.items(), memberships.items())
            ))
        else:
            raise Exception(f"Unknown pivot option, must match {PivotType}")

    def _process_query(
            self,
            phase: str,
            children_class: str,
            children: list,
            properties: list,
            samples: list,
            parent_class: str = "System",
            parent: str = "System",
            band: str = '1',
            period_type: str = "Interval",
            pivot: PivotType = "auto",
    ) -> pl.DataFrame:
        if self._horizon_type != (phase, period_type):
            self._horizon_type = (phase, period_type)
            self._horizon = query_horizon(self._pointer, phase, period_type)
        # Get all _id
        period_id = get_period_id(period_type)
        period_type_id = get_period_type_id(period_type)
        phase_id = get_phase_id(phase)
        timeslice_id = query_timeslice_id(self._pointer, "All Periods")
        # Membership dic
        membership_ids = self._kernel.findall_membership_id(parent_class, parent, children_class, children)
        memberships = dict(zip(children, membership_ids))
        # Property dic
        properties_name_id = {}
        for property_name in properties:
            property_unit = query_property_unit(self._pointer, property_name, children_class, parent_class)
            property_id = query_property_id(self._pointer, property_name, children_class, parent_class)
            properties_name_id[f"{property_name} ({property_unit})"] = property_id
        # Partial query
        array = self._array if period_type == "Interval" else self._yearray
        prefilter = self._kernel.findall_key(
            phase_id=phase_id,
            period_type_id=period_type_id,
            band_id=band,
            timeslice_id=timeslice_id,
        )
        return self._process_extraction(array, prefilter, period_id, memberships, properties_name_id, samples, pivot)

    def extract(
            self,
            phase: str,
            children_class: str,
            children: list,
            properties: list,
            samples: list,
            parent_class: str = "System",
            parent: str = "System",
            band: str = '1',
            period_type: str = "Interval",
            pivot: PivotType = "auto",
    ) -> pl.DataFrame:

        period_type = _get_period_type(period_type)

        if self._is_closed and period_type == "Interval":
            raise Exception("DarkSol is closed. No more hourly extraction.")
        elif period_type == "Interval" and self._array is None:
            raise QueryError("No Interval data to extract!")
        elif period_type == "FiscalYear" and self._yearray is None:
            raise QueryError("No FiscalYear data to extract!")

        try:
            df = self._process_query(
                phase,
                children_class,
                children,
                properties,
                samples,
                parent_class,
                parent,
                band,
                period_type,
                pivot
            )
            return df
        except (KeyError, StopIteration, QueryError) as extract_error:
            # Check inputs are valid
            assert phase in self.query_phase(), f"No {phase} data to extract!"
            assert parent_class in self.query_class(), f"'{parent_class}' not found"
            assert parent in self.query_children(parent_class), f"{parent_class} '{parent}' not found"
            assert children_class in self.query_class(), f"'{children_class}' not found"
            existing_children = self.query_children(children_class)
            for child in children:
                assert child in existing_children, f"{children_class} '{child}' not found"
            existing_properties = self.query_property(children_class, parent_class)
            for prop in properties:
                assert prop in existing_properties, f"{parent_class} {children_class} has no property '{prop}'"
            existing_samples = self.query_sample()
            for sample in samples:
                assert sample in existing_samples, f"Sample {sample} not found"
            # If all inputs are valid
            if not isinstance(extract_error, QueryError):
                raise extract_error
            query_error = f"{extract_error} in {phase}".replace("data", f"{period_type} data")
            raise QueryError(query_error)

    def query(
            self,
            phase: str,
            children_class: str,
            children: list,
            properties: list,
            samples: list,
            parent_class: str = "System",
            parent: str = "System",
            period_type: str = "Interval",
            pivot: PivotType = "auto",
    ) -> pd.DataFrame:
        """
        Perform query and extract result in pandas DataFrame.

        Parameters
        ----------
        phase
            Phase to query.
        children_class
            Class of the children. Must be the same for all children.
        children
            List of children to query.
        properties
            List of properties to query.
        samples
            List of samples to query.
        parent_class
            Class of the parent. Default is "System".
        parent
            Parent object of children. Default is "System".
        period_type
            Which PeriodType to query. Default is "Interval".
            Other available PeriodType is "FiscalYear".

            .. versionadded:: 0.1.3
                period_type is not case sensitive anymore (see below).

                - "Interval" can be written "interval" or "i"
                - "FiscalYear" can be written "Fiscal Year", "fiscalyear", "Year", "yearly" or "y"
        pivot
            Determines the pivot strategy used.
            "Pivot strategy" refers to the pivot() DataFrame method;
            it determines which result values should be added as a new column instead of stacked as new rows.

            - "auto" (default): Automatically switch between "none" and "property" strategy,
              depending which one will save the most rows ("property" strategy is prioritised).
            - "none": Each new child is added as a new column.
            - "property": Each property is added as a new column.
            - "melt": Put all result values in a "Value" column so the DataFrame stacks results only vertically.

        Returns
        -------
        A pandas DataFrame with ["Name", "Sample", "Datetime", "Property", "Value"] informations.
        The properties are renamed as "Property (Unit)".

        Examples
        --------
        >>> from fairyex import DarkSol
        >>>
        >>> with DarkSol("Model Open World Solution.zip") as ds:
        >>>     ds.query(
        ...         phase="STSchedule",
        ...         children_class="Generator",
        ...         children=["Limgrave foo"],
        ...         properties=["Generation"],
        ...         samples=["0", "1"],
        ...         # pivot as "auto" will be same as "property"
        ...     )
                     Name  Sample             Datetime  Generation (MW)
        0    Limgrave foo      0  01/01/2025 00:00:00              0.0
        1    Limgrave foo      1  01/01/2025 00:00:00              0.0
        >>>
        >>>     ds.query(
        ...         phase="STSchedule",
        ...         children_class="Generator",
        ...         children=["Limgrave foo", "Limgrave bar"],
        ...         properties=["Generation"],
        ...         samples=["0", "1"],
        ...         # pivot as "auto" will be same as "none"
        ...     )
                   Property Sample             Datetime  Limgrave foo  Limgrave bar
        0   Generation (MW)      0  01/01/2025 00:00:00           0.0           0.0
        1   Generation (MW)      1  01/01/2025 01:00:00           0.0           0.0
        """
        try:
            df = self.extract(
                phase,
                children_class,
                children,
                properties,
                samples,
                parent_class,
                parent,
                '1',
                period_type,
                pivot
            )
            return df.to_pandas()
        except (AssertionError, QueryError) as msg:
            query_error = QueryError(msg)
            logging.error("Query failed")
        raise query_error



from .querxml import (
    get_all, get_key_index, _query, _set_query
)
from .querid import (
    get_category_id, get_class_id, get_collection_id,
    get_membership_id, get_model_name, get_model_id, get_object_id,
    get_property_id, get_system_id, get_timeslice_id,
)


_PHASES = {
    # "Preview": '0',
    "LTPlan": '1',
    "PASA": '2',
    "MTSchedule": '3',
    "STSchedule": '4',
}


def query_horizon(xml, phase, period_type):
    if period_type == "Interval":
        return get_datetime(xml, phase)
    return list(get_fiscal_year(xml))


def get_fiscal_year(xml):
    return get_all(xml, "period_4", "year_ending", {})


def get_datetime(xml, phase):
    interval = {
        dt.find(_set_query(xml, "interval_id")).text: dt.find(_set_query(xml, "datetime")).text
        for dt in _query(xml, "period_0", {})
    }
    horizon = [interval[interval_id] for interval_id in _get_datetime(xml, phase)]
    return horizon


def _get_datetime(xml, phase):
    period_counter = {}
    phase_id = get_phase_id(phase)
    for period in _query(xml, f"phase_{phase_id}", {}):
        period_id = int(period.find(_set_query(xml, "period_id")).text)
        interval_id = int(period.find(_set_query(xml, "interval_id")).text)
        if current_interval_id := period_counter.get(period_id):
            period_counter[period_id] = min(current_interval_id, interval_id)
        else:
            period_counter[period_id] = interval_id
    for period_id in sorted(period_counter.keys()):
        yield str(period_counter[period_id])


def query_key_index(
        xml,
        parent,
        parent_class,
        child,
        child_class,
        model,
        phase,
        property_name,
        property_collection,
        period_type,
        band,
        sample,
        timeslice
):
    return get_key_index(
        xml,
        membership_id=query_membership_id(
            xml, parent_class, parent, child_class, child
        ),
        model_id=query_model_id(xml, model),
        phase_id=get_phase_id(phase),
        property_id=query_property_id(xml, property_name, property_collection),
        period_type_id=get_period_type_id(period_type),
        band_id=band,
        sample_id=sample,
        timeslice_id=query_timeslice_id(xml, timeslice),
    )


def query_collection_id(xml, child_class, parent_class):
    property_child_class_id = get_class_id(xml, child_class)
    property_parent_class_id = get_class_id(xml, parent_class)
    return get_collection_id(xml, property_child_class_id, property_parent_class_id)


def query_membership_id(
    xml,
    parent_class,
    parent,
    child_class,
    child,
):
    parent_class_id = get_class_id(xml, parent_class)
    parent_id = get_object_id(xml, parent, parent_class_id)
    child_class_id = get_class_id(xml, child_class)
    child_id = get_object_id(xml, child, child_class_id)
    return get_membership_id(
        xml, parent_id, parent_class_id, child_id, child_class_id,
    )


def query_model_name(xml, model_id='1'):
    return next(get_all(xml, "model", "name", {"model_id": model_id}))


def query_model_id(xml, model_name):
    return '1'


def get_phase_id(phase: str):
    return _PHASES[phase]


def get_period_id(period: str):
    return {"FiscalYear": '4', "Interval": '0'}[period]


def get_period_type_id(period_type: str):
    return {"FiscalYear": '1', "Interval": '0'}[period_type]


def query_property_id(xml, property_name, property_child_class, property_parent_class="System"):
    collection_id = query_collection_id(xml, property_child_class, property_parent_class)
    return get_property_id(xml, property_name, collection_id)


def query_property_unit(xml, property_name, property_child_class, property_parent_class="System"):
    property_child_class_id = get_class_id(xml, property_child_class)
    property_parent_class_id = get_class_id(xml, property_parent_class)
    collection_id = get_collection_id(xml, property_child_class_id, property_parent_class_id)
    property_filter = {"name": property_name, "collection_id": collection_id}
    unit_id = next(
        get_all(xml, "property", "unit_id", property_filter)
    )
    unit_value = next(
        get_all(xml, "unit", "value", {"unit_id": unit_id})
    )
    return unit_value


def query_timeslice_id(xml, timeslice_name):
    return get_timeslice_id(xml, timeslice_name)


def query_categories(xml, category_class):
    class_id = get_class_id(xml, category_class)
    return get_all(xml, "category", "name", {"class_id": class_id})


def query_classes(xml):
    return get_all(xml, "class", "name", {})


def query_children(xml, child_class, category=None):
    child_class_id = get_class_id(xml, child_class)
    children_filter = {"class_id": child_class_id}
    if category is not None:
        category_id = get_category_id(xml, category, child_class_id)
        children_filter |= {"category_id": category_id}
    return get_all(xml, "object", "name", children_filter)


def query_phases(xml):
    for phase_name, phase_id in _PHASES.items():
        if xml.find(_set_query(xml, f"t_phase_{phase_id}")):
            yield phase_name


def query_properties(xml, child_class, parent_class):
    collection_id = query_collection_id(xml, child_class, parent_class)
    return get_all(xml, "property", "name", {"collection_id": collection_id})


def query_samples(xml):
    return get_all(xml, "sample", "sample_id", {})


def query_samples_name(xml):
    return get_all(xml, "sample", "sample_name", {})



import unittest

from fairyex.darksol import _get_period_type, _get_sample_name


class TestDarkSolUtils(unittest.TestCase):

    def test_get_period_type(self):
        for period_type, user_period_types in {
            "Interval": ["Interval", "interval", "i"],
            "FiscalYear": ["FiscalYear", "Fiscal Year", "fiscalyear", "Year", "yearly", "y"],
        }.items():
            for user_period_type in user_period_types:
                self.assertEqual(_get_period_type(user_period_type), period_type)

    def test_get_sample_name(self):
        for sample_id, sample_name in {
            "-666": "Statistic 666",
            "-4": "Statistic 4",
            "-3": "Max",
            "-2": "Min",
            "-1": "Std",
            "0": "Mean",
            "1": "Sample 1",
            "42": "Sample 42",
        }.items():
            self.assertEqual(_get_sample_name(sample_id), sample_name)



"""QUERy on the XML file"""

from xml.etree import ElementTree


FLOAT_SIZE = 8


def _set_query(xml, query: str):
    root = xml.getroot()
    default_namespace = root.tag[:-len("SolutionDataset")]
    return f"{default_namespace}{query}"


def _query(xml, element: str, kwargs: dict):
    leaves = xml.findall(_set_query(xml, f"t_{element}"))
    for leaf in leaves:
        if all(
            leaf.find(_set_query(xml, filter_key)).text == filter_id
            for filter_key, filter_id in kwargs.items()
        ) or not kwargs:
            yield leaf


def get_all(xml: ElementTree, element: str, query: str, kwargs: dict):
    for leaf in _query(xml, element, kwargs):
        leaf_attr = leaf.find(_set_query(xml, query)).text
        yield leaf_attr


def _get_id(xml: ElementTree, element: str, **kwargs):
    for leaf in _query(xml, element, kwargs):
        id_found = leaf.find(_set_query(xml, f"{element}_id")).text
        return id_found


def _get_from_key_id(xml, element: str, key_name: str, key_id: str):
    leaves = xml.findall(_set_query(xml, f"t_{element}"))
    for leaf in leaves:
        if leaf.find(_set_query(xml, f"{key_name}_id")).text == key_id:
            return leaf


def get_key_index(xml: ElementTree, **kwargs):
    key_id = _get_id(xml, "key", **kwargs)
    leaf = _get_from_key_id(xml, "key_index", "key", key_id)
    position = leaf.find(_set_query(xml, "position")).text
    length = leaf.find(_set_query(xml, "length")).text
    return int(position) // FLOAT_SIZE, int(length)



"""QUERy based on ID """

from .querxml import  _get_id


def get_system_id():
    return '1'


def get_category_id(xml, category_name, category_class_id):
    return _get_id(xml, "category", name=category_name, class_id=category_class_id)


def get_class_id(xml, class_name):
    return _get_id(xml, "class", name=class_name)


def get_collection_id(xml, child_class_id, parent_class_id):
    return _get_id(
        xml, "collection", child_class_id=child_class_id, parent_class_id=parent_class_id
    )


def get_membership_id(
    xml,
    parent_id,
    parent_class_id,
    child_id,
    child_class_id,
):
    return _get_id(
        xml,
        "membership",
        parent_object_id=parent_id,
        parent_class_id=parent_class_id,
        child_object_id=child_id,
        child_class_id=child_class_id,
    )


def get_model_name(xml, model_id):
    return _get_id(xml, "model", model_id=model_id)


def get_model_id(xml, model_name):
    return _get_id(xml, "model", name=model_name)


def get_object_id(xml, object_name, class_id):
   return _get_id(
        xml, "object", name=object_name, class_id=class_id
    )


def get_property_id(xml, property_name, collection_id):
    return _get_id(
        xml, "property", name=property_name, collection_id=collection_id
    )


def get_timeslice_id(xml, timeslice_name):
    return _get_id(xml, "timeslice", name=timeslice_name)
